"""Simple Graphics

This module allows the user to open a monochrome graphics window and draw (and
undraw) squares in that window. It does *not* handle user events. The colors
are fixed to have a black background and a "classic monochrome" green
foreground.

Basic usage:

    from simple_graphics import SimpleGraphics

    sg = SimpleGraphics()    # opens graphics window
    sg.paint(0, 0)           # draws box in upper-left corner
    w = sg.width             # number of columns in the window
    h = sg.height            # number of rows
    sg.paint(w - 1, h - 1)   # draws box in lower-right corner

Defines class SimpleGraphics which upon construction opens a window.
SimpleGraphics objects offer these attributes:

width:  the number of columns (of boxes, not pixels) in the window
height: the number of rows

and these methods:

paint(x, y):       draws green square at column x, row y
unpaint(x, y):     removes square at x, y if one is there (paints it black)
count_squares():   returns the number of green squares currently on screen
is_painted(x, y):  returns True if there is x, y is currently painted
set_status(s):     replaces text on status bar with string s
close():           closes the graphics window

Known bugs:

Due to the underlying graphics engine (TkInter), sometimes a graphics window
can be closed - but the window will not visibly close on screen. If that
happens, when a new window is constructed, the old one will close.

"""


from sys import stderr
import tkinter as tk
import warnings


MIN_ROWS = 1
DEFAULT_ROWS = 25
MIN_COLUMNS = MIN_ROWS
DEFAULT_COLUMNS = 40
MIN_SCALE = 3
DEFAULT_SCALE = 20
MAX_ROWS = 1000 // MIN_SCALE
MAX_COLUMNS = MAX_ROWS
DEFAULT_BG_COLOR = 'black'

PAD = 5   # pixel width of "highlight thickness" of canvas


class _SquareGraphics:
    def __init__(self,
                 columns=DEFAULT_COLUMNS,
                 rows=DEFAULT_ROWS,
                 scale=DEFAULT_SCALE,
                 bg_color=DEFAULT_BG_COLOR,
                 init_fg_color=None,
                 show_cell_borders=False,
                 force_update=False,
                 title=''):
        root = tk.Tk()

        cell_border = 1 if show_cell_borders else 0
        screen_width = root.winfo_screenwidth() - 2*PAD
        screen_height = root.winfo_screenheight() - 2*PAD
        screen_width -= cell_border
        screen_height -= cell_border

        if rows < MIN_ROWS or rows > MAX_ROWS:
            warnings.warn(f'rows ({rows}) out of range - ' +
                          f'using default ({DEFAULT_ROWS})')
            rows = DEFAULT_ROWS
        if columns < MIN_COLUMNS or rows > MAX_COLUMNS:
            warnings.warn(f'columns ({columns}) out of range - ' +
                          f'using default ({DEFAULT_COLUMNS})')
            columns = DEFAULT_COLUMNS
        if scale < MIN_SCALE:
            warnings.warn(f'scale ({scale}) too small - ' +
                          f'using default ({DEFAULT_SCALE})')
        width = scale * columns
        if screen_width <= width:
            scale = screen_width // columns
            width = scale * columns
            warnings.warn('scale too large for screen width, adjusting')
        height = scale * rows
        if screen_height <= height:
            scale = screen_height // rows
            height = scale * rows
            warnings.warn('scale too large for screen height, adjusting')
        width += cell_border
        height += cell_border

        border_color = _gray_complement(root, bg_color)
        if not init_fg_color:
            init_fg_color = _gray_complement(root, bg_color)

        self._root = root
        self._width = width
        self._height = height
        self._rows = rows
        self._columns = columns
        self._scale = scale
        self._bg_color = bg_color
        self._border_color = border_color
        self._init_fg_color = init_fg_color
        self._cell_border = cell_border
        self._force_update = force_update
        self._title = title

        self._initialize()

    def _initialize(self):
        self._root.title(self._title)
        self._root.wm_protocol('WM_DELETE_WINDOW', self.close)

        canvas = tk.Canvas(self._root,
                           width=self._width,
                           height=self._height,
                           bg=self._bg_color,
                           highlightthickness=PAD,
                           highlightcolor=self._border_color)
        square_map = {}
        for y in range(self._rows):
            for x in range(self._columns):
                left = PAD + x * self._scale
                top = PAD + y * self._scale
                nw = (left, top)
                se = (left + self._scale, top + self._scale)
                rid = canvas.create_rectangle(nw, se,
                                              fill=self._bg_color,
                                              width=self._cell_border)
                square_map[(x, y)] = rid
        canvas.pack(side=tk.TOP)

        status_bar_frame = tk.Frame(self._root)
        status_bar_frame.pack(side=tk.BOTTOM, fill=tk.X)
        status_bar = tk.Label(status_bar_frame,
                              font='TkFixedFont',
                              pady=PAD, padx=PAD)   # !!! check this
        status_bar.pack(fill=tk.X)

        self._canvas = canvas
        self._status_bar = status_bar
        self._last_color = self._init_fg_color
        self._square_map = square_map
        self._active_squares = {}

    def restart(self):
        if self._root is not None and len(self._root.children) > 0:
            self.close()
        self._root = tk.Tk()
        self._initialize()

    def _validate(self, p=None):
        if self._root is None or len(self._root.children) == 0:
            raise SquareGraphicsException('no active graphics window')

    def _is_valid_point(self, p):
        (x, y) = p
        return (isinstance(x, int) and isinstance(y, int) and
                x >= 0 and x < self._columns and y >= 0 and y < self._rows)

    def close(self):
        """Closes graphics window."""
        self._validate()
        self._root.destroy()
        self._root = None

    @property
    def width(self):
        return self._columns

    @property
    def height(self):
        return self._rows

    def get_dimensions(self):
        """Return dimensions of opened graphics windows columns by rows."""
        self._validate()
        return (self._columns, self._rows)

    def paint(self, p, color=None):
        """Paints cell at position p (a column-row pair) to be specified color.
        """
        self._validate()
        if self._is_valid_point(p):
            if color is None:
                color = self._last_color
            if not color == self._bg_color:
                self._last_color = color
            self._canvas.itemconfig(self._square_map[p], fill=color)
            if not color == self._bg_color:
                self._active_squares[p] = color
            else:
                self._active_squares.pop(p, None)
            if self._force_update:
                self._root.update()
        else:
            warnings.warn(f'ignoring invalid coordinates {p}')

    def unpaint(self, p):
        """
        Clears cell at position p (a column-row pair).
        Equivalent to painting cell with background color.
        """
        self._validate()
        if self._is_valid_point(p):
            if p in self._active_squares:
                self._canvas.itemconfig(self._square_map[p],
                                        fill=self._bg_color)
                self._active_squares.pop(p, None)
                if self._force_update:
                    self._root.update()
        else:
            warnings.warn(f'ignoring invalid coordinates {p}')

    def clear(self):
        """Clears graphics window. (All cells reset to background color.)"""
        self. _validate()
        for p in self._active_squares.keys():
            self._canvas.itemconfig(self._square_map[p], fill=self._bg_color)
        self._active_squares.clear()
        if self._force_update:
            self._root.update()

    def set_status(self, message, justify='left'):
        """Update message in status bar at bottom of graphics window."""
        self._validate()
        match justify:
            case 'right':  anchor = tk.E
            case 'center': anchor = tk.CENTER
            case 'left':   anchor = tk.W
            case _:
                anchor = tk.W
                warnings.warn(f'unexpected justify "{justify}"')
        self._status_bar.config(text=message, anchor=anchor)
        if self._force_update:
            self._root.update()


class SquareGraphicsException(Exception):
    pass


def _gray_complement(tk_widget, color):
    # produces Tkinter color (either as hex #RRRRGGGGBBBB or name)
    # that is gray (r==g==b) and "complementary" (white, black, gray) to
    # given color
    mx = 1 << 16   # Tkinter uses 16 bits for each channel
    half = mx / 2
    quarter = half / 2
    avg = sum(tk_widget.winfo_rgb(color)) / 3
    if avg < quarter or avg > mx - quarter:
        return 'gray'
    elif avg < half:
        return 'white'
    else:
        return 'black'


def _show_warning(message, *ignore_args):
    print(message, file=stderr, flush=True)


warnings.showwarning = _show_warning


# ------------------------------------------------------------------------


_BG_COLOR = 'black'
_FG_COLOR = '#3FFFBBBB3FFF'
_ROWS = 25
_COLUMNS = 40
_SCALE = 15


class SimpleGraphics(_SquareGraphics):
    def __init__(self):
        super().__init__(rows=_ROWS,
                         columns=_COLUMNS,
                         scale=_SCALE,
                         bg_color=_BG_COLOR,
                         init_fg_color=_FG_COLOR,
                         force_update=True,
                         title='Simple Graphics')

    def is_painted(self, x, y):
        self._validate()
        return (x, y) in self._active_squares

    def count_squares(self):
        self._validate()
        return len(self._active_squares)

    def paint(self, x, y):
        super().paint((x, y))

    def unpaint(self, x, y):
        super().unpaint((x, y))
